/* LICENSE:
Copyright (c) Members of the EGEE Collaboration. 2010.
See http://www.eu-egee.org/partners/ for details on the copyright
holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied.
See the License for the specific language governing permissions and
limitations under the License.

END LICENSE */

#include <map>
#include <istream>
#include <fstream>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <getopt.h>

#include "glite/ce/cream-client-api-c/CreamProxyFactory.h"
#include "glite/ce/cream-client-api-c/JobFilterWrapper.h"
#include "glite/ce/cream-client-api-c/ResultWrapper.h"
#include "glite/ce/cream-client-api-c/certUtil.h"
#include "glite/ce/cream-client-api-c/ConfigurationManager.h"
#include "glite/ce/cream-client-api-c/creamApiLogger.h"
#include "glite/ce/cream-client-api-c/VOMSWrapper.h"
#include "glite/ce/cream-client-api-c/CEUrl.h"
#include "util/cliUtils.h"
#include "hmessages.h"
#include "boost/algorithm/string.hpp"
#include "boost/format.hpp"

using namespace std;
using namespace glite::ce::cream_client_api::soap_proxy;
using namespace glite::ce::cream_client_api::util;
using namespace glite::ce::cream_client_api;
using namespace glite::ce::cream_client_api::cream_exceptions;

Namespace namespaces[] = {};

void printhelp(void) ;

int main(int argc, char *argv[]) {

  string jobID;
  string endpoint;
  bool debug = false;
  bool hostset=false;
  char c;
  int option_index = 0;
  string logfile;
  bool redir_out = false;
  bool joblist_from_file=false;
  string joblist_file = "";
  
  bool nomsg = false;
  bool noint = false;
  bool all = false;
  vector<string> all_jobids_to_resume, all_jobids_to_actually_resume;
  map <string, vector<string> > url_jobids;
  bool resume_all_jobs_from_file = false;
  string    user_conf_file = "";
  string certfile;
  string proxyfile;
  bool specified_proxyfile=false;

  int timeout = 30;

  creamApiLogger* logger_instance = creamApiLogger::instance();
  log4cpp::Category* log_dev = logger_instance->getLogger();

  // Initialize the logger
  log_dev->setPriority( log4cpp::Priority::NOTICE );
  logger_instance->setLogfileEnabled( true );

  all_jobids_to_resume.reserve(100);
  all_jobids_to_actually_resume.reserve(100);

  bool verify_ac_sign = true;

  while(1) {
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"debug", 0, 0, 'd'},
      {"version", 0, 0, 'v'},
      {"endpoint", 1, 0, 'e'},
      {"logfile", 1, 0, 'l'},
      {"input", 1, 0, 'i'},
      {"noint", 0, 0, 'N'},
      {"all", 0, 0, 'a'},
      {"nomsg", 0, 0, 'n'},
      {"conf", 1, 0, 'c'},
      {"proxyfile", 1, 0, 'p'},
      {"timeout", 1, 0, 't'},
      {"donot-verify-ac-sign", 0, 0, 'A'},
      {0, 0, 0, 0}
    };
    c = getopt_long(argc, argv, "p:l:nAahdve:i:Nc:t:", long_options, &option_index);
    if ( c == -1 )
      break;

    switch(c) {
    case 'A':
      verify_ac_sign = false;
      break;
    case 'p':
      proxyfile = string(optarg);
      specified_proxyfile = true;
      break;
    case 'a':
      all=true;
      break;
    case 'N':
      noint=true;
      break;
    case 'v':
      help_messages::printver();
      exit(0);
      break;
    case 'h':
      printhelp();
      exit(0);
      break;
    case 'd':
      debug=true;
      log_dev->setPriority( log4cpp::Priority::DEBUG );
      break;
    case 'l':
      redir_out = true;
      logfile = string(optarg);
      break;
    case 'n':
      nomsg = true;
      log_dev->setPriority( log4cpp::Priority::ERROR );
      break;
    case 'i':
      joblist_from_file=true;
      joblist_file = string(optarg);
      break;
    case 'e':
      endpoint = string(optarg);
      hostset=true;
      break;
    case 'c':
      user_conf_file = string(optarg);
      break;
    case 't':
      timeout = atoi(optarg);
      break;
    default:
      cerr << "Type " << argv[0] << " -h for an help"<<endl;
      exit(1);

    }
  }

  if(nomsg && debug) {
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, "Cannot specify both --debug and --nomsg options. Stop.", false, true, true);
    exit(1);
  }

  AbsCreamProxy *creamClient;
  string VO = "";

  if(!specified_proxyfile) {
    try {
      certfile = cliUtils::getProxyCertificateFile();
    } catch(exception& ex) {
      logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
      exit(1);
    }
  } else
    certfile = proxyfile;
  
  if(debug)
    logger_instance->log(log4cpp::Priority::INFO, string("Using certificate proxy file [")
			 + certfile + "]", true, true, true);

//       long int leftcert = certUtil::getProxyTimeLeft(certfile);

//       if(leftcert <= 0) {
//         log_dev->fatal( "Proxy certificate is expired. Please Check it." );
// 	exit(1);
//       }

//     VO = certUtil::getVOName( certfile );

    VOMSWrapper V( certfile, verify_ac_sign  );
    if( !V.IsValid( ) ) {
      
	logger_instance->log(log4cpp::Priority::FATAL, 
			     string("Problems with proxyfile [")+certfile+"]: "+
			     V.getErrorMessage(), true, true, true);
	exit(1);
      
    }
//       } else
//       logger_instance->log(log4cpp::Priority::ERROR, 
// 			   "Error while reading proxyfile ["+certfile+"]: "+
// 			   V.getErrorMessage(), true, true, true);
// 	exit(1);
//     }
    
//     long int leftcert = V.getProxyTimeEnd( ) - time( NULL );
    
//     if(leftcert <= 0) {
//       log_dev->fatal( "Proxy certificate is expired. Please Check it." );
//       exit(1);
//     }
    
    VO = V.getVOName( );

    /**
     * if all is ON 'all_jobids_to_resume' is an empty vector
     */

  
  /**
   * Handle the load of configurations
   */
  //  vector<string> confiles = cliUtils::getConfigurationFiles(VO,
  //							    user_conf_file, 
  //							    DEFAULT_CONF_FILE);

  vector<string> confiles;
  try {
    confiles = cliUtils::getConfigurationFiles(VO,
					       user_conf_file, 
					       DEFAULT_CONF_FILE);
  } catch(no_user_confile_ex& ex) {
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  }

  if(debug) {
    for(vector<string>::const_iterator fit = confiles.begin();
        fit != confiles.end();
        fit++) //unsigned int j=0; j<confiles.size(); j++)
      logger_instance->log(log4cpp::Priority::INFO, 
			   string("Loading configurations from [")
			   + *fit +"]...", true, true, true);
  }

  ConfigurationManager confMgr(ConfigurationManager::classad);
  try { confMgr.load(confiles); }
  catch(file_ex& ex) {
    if(debug)
    logger_instance->log(log4cpp::Priority::WARN, 
			   string(ex.what())+". Using built-in configuration",
			   true,true,!nomsg);
  }

  if(debug || redir_out) {
    if(!redir_out)
      logfile = logger_instance->getLogFileName(confMgr.getProperty("RESUME_LOG_DIR", DEFAULT_LOG_DIR).c_str(), "glite-ce-job-resume");
    cliUtils::mkdir(cliUtils::getPath(logfile));
    logger_instance->setLogFile(logfile.c_str());
  }
  if(debug) {
    string os = string("Logfile is [") + logfile + "]";

    logger_instance->log(log4cpp::Priority::INFO, os, false, false, !nomsg);
  }

  if( (all && joblist_from_file) || (argv[optind]!=NULL && all)){
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, "--all and --input or --all and specification of JobID(s) as argument are exclusive. Stop.", true, true, true);
    exit(1);
  }

  if( (argv[optind]==NULL) && (!joblist_from_file) && (!all)) {
    //ERROR
    logger_instance->log(log4cpp::Priority::FATAL,"No Job ID nor JobList file have been specified in the command line arguments. Stop.", false, true, true);
    exit(1);
  } 
  
  if(all && !hostset) {
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, "Option --all requires the specification of the endpoint (option --endpoint) to contact. Stop.", true, true, true);
    exit(1);
  }
  
  if(debug) {
    logger_instance->log(log4cpp::Priority::INFO, "*************************************", false, true, !nomsg);
    logger_instance->log(log4cpp::Priority::INFO, help_messages::getStartMessage().c_str(), false, true, !nomsg);
  }

  if(hostset) {

    if( !cliUtils::checkEndpointFormat( endpoint ) )
    {
      log_dev->fatal( "Endpoint not specified in the right format: should be <host>[:tcpport]. Stop.");
      exit(1);
    }

//     boost::regex pattern( "^[^:]+(:[0-9]{1,5})?\\b" );
//     boost::smatch what;
//     if( !boost::regex_match( endpoint, what, pattern ) ) {
//     //if(!string_manipulation::matches(endpoint,"^[^:]+(:[0-9]{1,5})?\\b")) {
//       logger_instance->log(log4cpp::Priority::ERROR,"Specified endpoint has wrong format. Must be <host>:<tcp_port> (tcp_port is a number of 5 digit max length). Stop", true, true,true);
//       exit(1);
//     }

  }

  /**
   * The following piece of code is executed only if --all hasn't been specified
   *
   */
  while(argv[optind]!=NULL) {
    all_jobids_to_resume.push_back(string(argv[optind]));
    optind++;
  }

  /**
   *
   * Gets job IDs from input file and asks the user
   * This piece of code is not executed if all is ON (in
   * fact --all disallows --input specification)
   *
   */
  string serviceAddress = "";
  unsigned int jcounter = 0;
  vector<string> jobnumbers_to_resume;
  jobnumbers_to_resume.reserve(100);
  if(joblist_from_file) {

    string errMessage;
    
    if(!cliUtils::interactiveChoice("Resume", joblist_file.c_str(), noint, debug, nomsg, resume_all_jobs_from_file, jobnumbers_to_resume, all_jobids_to_resume, errMessage)) {
      logger_instance->log(log4cpp::Priority::FATAL, errMessage, true, true, true);
      exit(1);
    }

//     bool isjoblist = true;
//     int f;
//     if(debug) {
//       string os = string("Getting job list from file [") + joblist_file + "]";
//       logger_instance->log(log4cpp::Priority::INFO, os, true, true, !nomsg);
//     }
    
//     if((f=open(joblist_file.c_str(), O_RDONLY))==-1) {
//       int saveerr = errno;
//       string os = string("Error accessing file [") + joblist_file + "]: " + strerror(saveerr) + ". Stop.";
//       logger_instance->log(log4cpp::Priority::FATAL, os, true, true, true);
//       exit(1);
//     }
//     close(f);
    
//     isjoblist = cliUtils::isACreamJobListFile(joblist_file.c_str());
//     if(!isjoblist) {
//       string os = string("File [") + joblist_file + "] is not a CREAM job list file. Stop.";
//       logger_instance->log(log4cpp::Priority::FATAL, os, true,true, true);
//       exit(1);
//     }
//     cliUtils::getJobIDFromFile(all_jobids_to_resume, joblist_file.c_str());
    
//     if(!noint) {
//       string os = string("\nFound the following jobs in [") + joblist_file + "]: ";
//       logger_instance->log(log4cpp::Priority::INFO, os, false, false, true);
//     }
//     string whichone = "";
//     if(!noint) {
//       printf("\n");
//       for(vector<string>::const_iterator jit = all_jobids_to_resume.begin();
//           jit != all_jobids_to_resume.end();
//           jit++) { //jcounter=0; jcounter<all_jobids_to_resume.size(); jcounter++) {
// //        string os;
// //	os = os + boost::str( boost::format("%d") % jcounter) + ". " + *jit;//all_jobids_to_resume.at(jcounter);
//         printf("%s\n", (boost::str( boost::format("%d") % jcounter++) + ". " + *jit).c_str());
//       }
//       printf("a. Resume all jobs\n");
//       printf("q. Quit\n");
//       printf("\nYour choice ? (multiple choices separated by ',') ");
//       cin >> (string&)whichone;
//       if( whichone == "q" ) {
//         printf("Resume aborted. Bye.\n");
//         exit(0);
//       }

      
//       /**
//        * must check the format of whichone
//        */
//       if(whichone != "a") {
// 	/**
// 	 * check for a format like \d[,\d]+
// 	 */
// 	boost::cmatch what;
// 	const boost::regex pattern("[0-9]+(,[0-9]+)*");
// 	if(!boost::regex_match(whichone.c_str(), what, pattern) ) {
// 	  logger_instance->log(log4cpp::Priority::FATAL, "Response must be \"a\" or \"q\" or numbers ','-separated. Stop", false, false, true);
// 	  exit(1);
// 	}
//       }

//     } else whichone = "a";
//     if( whichone == "a" ) resume_all_jobs_from_file=true;
//     boost::split(jobnumbers_to_resume, whichone, boost::is_any_of(","));//, jobnumbers_to_resume );
  }

  /**
   * Add the tcpport if not specified on the command line
   */
//   boost::regex pattern( "^\\w+(:\\d+)" );
//   boost::smatch _what;
//   //printf("endpoint=[%s]\n", endpoint.c_str());
//   if( !boost::regex_match( endpoint, _what, pattern ) ) {
  if(!cliUtils::containsTCPPort( endpoint )) {
    endpoint = endpoint + ":" + confMgr.getProperty("DEFAULT_CREAM_TCPPORT", "8443");
    //printf("endpoint=[%s]\n", endpoint.c_str());
  }
  if(!all) {
    /**
     *
     * Asks confirmation about resume of selected job IDs
     *
     */
    char confirm;
    if(!noint) {
      printf("\nAre you sure you want to resume specified job(s) [y/n]: ");
      cin >> confirm;
    } else confirm ='y';

    if(confirm != 'y') {
      printf("Resume aborted. Bye.\n");
      exit(0);
    }

    /**
     *
     * Filters out the job IDs the user didn't select
     *
     */
    for(jcounter=0; jcounter<all_jobids_to_resume.size(); jcounter++) {
      if(joblist_from_file) {
        if(!resume_all_jobs_from_file) {
	  bool isthere = false;
	  for(
	      vector<string>::iterator it = jobnumbers_to_resume.begin();
	      it!=jobnumbers_to_resume.end();
	      it++)
	    if( jcounter == atoi( (*it).c_str()) ) { isthere = true; break; }
	  if(!isthere) continue;
        }
	if(debug)
	  logger_instance->log(log4cpp::Priority::INFO, string("Will resume job [")+all_jobids_to_resume.at(jcounter)+"]", true, true, true);
	all_jobids_to_actually_resume.push_back(all_jobids_to_resume.at(jcounter));
      }
      else all_jobids_to_actually_resume.push_back(all_jobids_to_resume.at(jcounter));
    }
    /**
     *
     * Create an hash map CEId -> JobIDs
     *
     */
    vector<string> pieces;
    pieces.reserve(10);

    for(vector<string>::const_iterator jit = all_jobids_to_actually_resume.begin();
	jit != all_jobids_to_actually_resume.end();
	jit++) { //unsigned int j=0; j<all_jobids_to_actually_resume.size(); j++) {
      if(hostset) {
	string tmp = confMgr.getProperty("CREAM_URL_PREFIX", DEFAULT_PROTO)+
	  endpoint+"/"+confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX);
        url_jobids[tmp].push_back( /*all_jobids_to_actually_resume.at(j)*/ *jit );
      }
      else {
        pieces.clear();
	//	pieces.reserve(10);
        try { CEUrl::parseJobID(/*all_jobids_to_actually_resume.at(j)*/ *jit, pieces, confMgr.getProperty("DEFAULT_CREAM_TCPPORT", "8443")); }
        catch(CEUrl::ceid_syntax_ex& ex) {
          logger_instance->log(log4cpp::Priority::ERROR, ex.what(), true, true, true);
          logger_instance->log(log4cpp::Priority::ERROR,"This JobID doesn't contain the CREAM service address. Skipping it...",true, true, true);
	  continue;
        }
	string tmp = confMgr.getProperty("CREAM_URL_PREFIX", DEFAULT_PROTO)+
	  pieces[1]+":"+pieces[2]+"/"+
	  confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX);
	 url_jobids[tmp].push_back( *jit /*all_jobids_to_actually_resume.at(j)*/ );
      }
    }
  } else {
    serviceAddress = confMgr.getProperty("CREAM_URL_PREFIX", DEFAULT_PROTO);
    serviceAddress = serviceAddress + endpoint + "/" + confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX);
  }

  JobFilterWrapper *wr;
  ResultWrapper result;

  try {
    if(all && !noint) {
      char confirm;
      printf("\nGoing to resume all job from CE [%s]. Do you want to proceed [y/n]? ",
	     serviceAddress.c_str());
      cin >> confirm;
      if(confirm!='y') {
        printf("Resume aborted. Bye.\n\n");
        exit(0);
      }
    }
    if(all) {
      vector<string> empty;
      logger_instance->log(log4cpp::Priority::INFO, string("Resuming all jobs on [")+serviceAddress+"]...", true, true, true);
      //initSoap->Resume(serviceAddress.c_str(), empty );

      wr = new JobFilterWrapper(vector<JobIdWrapper>(), vector<string>(), -1, -1, "", "");
      creamClient = CreamProxyFactory::make_CreamProxyResume( wr, &result, timeout );
      if(!creamClient)
	{
	  logger_instance->log(log4cpp::Priority::FATAL, "FAILED CREATION OF AN AbsCreamProxy object! STOP!!", true, true, true);
	  return 1;
	}
      boost::scoped_ptr< AbsCreamProxy > tmpClient;
      tmpClient.reset( creamClient );
      creamClient->setCredential( certfile );
      creamClient->execute( serviceAddress );
      cliUtils::printResult( result );
      //creamClient->garbage_collect();
      delete wr;
    }
    else {
      for(map<string, vector<string> >::iterator it=url_jobids.begin(); it!=url_jobids.end(); it++) {
	if(debug)
	  logger_instance->log(log4cpp::Priority::INFO, string("Resuming selected jobs on [")+(*it).first+"]...", true, true, true);
	//initSoap->Resume((*it).first.c_str(), (*it).second);
	
	vector<JobIdWrapper> target;
	stripCreamURL strip( &target, &confMgr );
	for_each(it->second.begin(), it->second.end(), strip);
	wr = new JobFilterWrapper(target, vector<string>(), -1, -1, "", "");       
	creamClient = CreamProxyFactory::make_CreamProxyResume( wr, &result, timeout );
	if(!creamClient)
	  {
	    logger_instance->log(log4cpp::Priority::FATAL, "FAILED CREATION OF AN AbsCreamProxy object! STOP!!", true, true, true);
	    return 1;
	  }
	boost::scoped_ptr< AbsCreamProxy > tmpClient;
	tmpClient.reset( creamClient );
	creamClient->setCredential( certfile );
	creamClient->execute( (*it).first.c_str() );
	cliUtils::printResult( result );
	//creamClient->garbage_collect();
	delete wr;
      }
    }
  } catch(BaseException& ex ) {
      log_dev->log(log4cpp::Priority::ERROR, ex.what(), true, true, true);
      exit(1);
  } catch(exception& s) {
    logger_instance->log(log4cpp::Priority::FATAL, s.what(), true, true, true );
    exit(1);
  } /*catch(InternalException& ex) {
    logger_instance->log(log4cpp::Priority::ERROR, ex.what(), true, true, true  );
    exit(1);
  }*/
  return 0;
}

void printhelp() {
  printf("%s\n",help_messages::getResumeHelp().c_str());
}
