/* LICENSE:
Copyright (c) Members of the EGEE Collaboration. 2010.
See http://www.eu-egee.org/partners/ for details on the copyright
holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied.
See the License for the specific language governing permissions and
limitations under the License.

END LICENSE */

/**
 *
 * SYSTEM C/C++ INCLUDE FILES
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <strings.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <cstring>
#include <fcntl.h>
#include <cstdlib>
#include <fstream>
#include <dlfcn.h>
#include <cstdio>

/**
 *
 * LOCAL INCLUDE FILES
 *
 */
#include "ftpclient/gridftpClient.h"
#include "util/jdlHelper.h"
#include "util/cliUtils.h"
#include "hmessages.h"

/**
 *
 * CREAM CLIENT API C++ INCLUDE FILES
 *
 */
#include "glite/ce/cream-client-api-c/JobDescriptionWrapper.h"
#include "glite/ce/cream-client-api-c/ConfigurationManager.h"
#include "glite/ce/cream-client-api-c/JobPropertyWrapper.h"
#include "glite/ce/cream-client-api-c/CreamProxyFactory.h"
#include "glite/ce/cream-client-api-c/JobFilterWrapper.h"
#include "glite/ce/cream-client-api-c/CreamProxy_Impl.h"
#include "glite/ce/cream-client-api-c/creamApiLogger.h"
#include "glite/ce/cream-client-api-c/ResultWrapper.h"
#include "glite/ce/cream-client-api-c/JobIdWrapper.h"
#include "glite/ce/cream-client-api-c/VOMSWrapper.h"
#include "glite/ce/cream-client-api-c/certUtil.h"
#include "glite/ce/cream-client-api-c/file_ex.h"
#include "glite/ce/cream-client-api-c/CEUrl.h"

/**
 *
 * BOOST INCLUDE FILES
 *
 */
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include "boost/algorithm/string.hpp"
#include "boost/tuple/tuple.hpp"
#include "boost/format.hpp"
#include "boost/regex.hpp"

#define MATCH_NUM 10
#define DEFAULT_GRIDFTP_CLIENT "globus-url-copy -p "
#define DEFAULT_CEURLDELEGATION_POSTFIX "/ce-cream/services/gridsite-delegation"


using namespace std;
using namespace glite::ce::cream_client_api;
using namespace glite::ce::cream_client_api::util;
using namespace glite::ce::cream_client_api::cream_exceptions;
using namespace glite::ce::cream_client_api::soap_proxy;

namespace fs = boost::filesystem;

Namespace namespaces[] = {};

void printhelp();

string make_delegation( const string&, const string &, const int ) throw();

int main(int argc, char *argv[]) {

  string URL;
  string JDLFile;
  string VO_from_cert = "";
  bool jdlset = false;
  bool mustcheckVO = false;
  bool debug = false;
  bool ceidset = false;
  bool automatic_delegation = false;

  string    logfile;
  string    outfile;
  bool      jobid_on_file = false;
  int       option_index = 0;
  bool      cli_specified_vo = false;
  string    user_VO;
  bool      nomsg = false;
  bool      fake = false;
  bool      specifiedDID = false;
  string    DID;
  ofstream  filestr;
  bool      jobmpi = false;
  bool      noint = false;
  int       num_streams = 4;
  bool      redir_out = false;
  string    user_conf_file = "";
  creamApiLogger* logger_instance = creamApiLogger::instance();
  log4cpp::Category* log_dev = logger_instance->getLogger();
  string certfile, proxyfile;
  bool specified_proxyfile = false;
  string leaseId = "";

  int timeout = 30;

  // Initialize the logger
  log_dev->setPriority( log4cpp::Priority::NOTICE );
  logger_instance->setLogfileEnabled( true );

    bool verify_ac_sign = true;

  /* ****************************************************************************
   * Will parse all options
   */ 
  while(1) {
    int c;
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"debug", 0, 0, 'd'},
      {"version", 0, 0, 'v'},
      {"nomsg", 0, 0, 'n'},
      {"fake", 0, 0, 'F'},
      {"resource", 1, 0, 'r'},
      {"logfile", 1, 0, 'l'},
      {"output", 1, 0, 'o'},
      {"autm-delegation", 0, 0, 'a'},
      {"delegationId", 1, 0, 'D'},
      {"noint", 0, 0, 'N'},
      {"vo", 1, 0, 'V'},
      {"ftp-streams", 1, 0, 's'},
      {"conf", 1, 0, 'c'},
      {"proxyfile", 1, 0, 'p'},
      {"leaseId", 1, 0, 'L'},
      {"timeout", 1, 0, 't'},
      {"donot-verify-ac-sign", 0, 0, 'A'},
      {0, 0, 0, 0}
    };

    c = getopt_long(argc, argv, "D:hdAvr:o:l:aV:FnNs:c:p:L:t:", long_options, &option_index);
    if ( c == -1 )
      break;

    switch(c) {
    case 'A':
      verify_ac_sign = false;
      break;      
    case 'p':
      proxyfile = string(optarg);
      specified_proxyfile = true;
      break;
    case 'h':
      printhelp();
      exit(0);
      break;
    case 'd':
      debug=true;
      log_dev->setPriority( log4cpp::Priority::DEBUG );
      break;
    case 'N':
      noint = true;
      break;
    case 'D':
      specifiedDID = true;
      DID = string(optarg);
      break;
    case 'a':
      automatic_delegation=true;
      break;
    case 'v':
      help_messages::printver();
      exit(0);
      break;
    case 's':
      num_streams = atoi(optarg);
      break;
    case 'l':
      redir_out = true;
      logfile = string(optarg);
      break;
    case 'V':
      user_VO = string(optarg);
      cli_specified_vo = true;
      break;
    case 'o':
      jobid_on_file = true;
      outfile = string(optarg);
      break;
    case 'n':
      nomsg = true;
      log_dev->setPriority(log4cpp::Priority::ERROR );
      break;
    case 'F':
      fake = true;
      break;
    case 'r':
      URL = string(optarg);
      ceidset=true;
      break;
    case 'c':
      user_conf_file = string(optarg);
      break;
    case 'L':
      leaseId = string(optarg);
      break;
    case 't':
      timeout = atoi(optarg);
      break;
    default:
      cerr << "Type " << argv[0] << " -h for an help"<<endl;
      exit(1);
    }
  }
  
  /* ****************************************************************************
   * Will check all specified options
   */
  if(nomsg && debug) {
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL,
			 "Cannot specify both --debug and --nomsg options. Stop.", 
			 false, 
			 true, 
			 true);
    exit(1);
  }
 
  if(debug) {
    logger_instance->log(log4cpp::Priority::INFO, 
			 "*************************************", 
			 false, true, debug);

    logger_instance->log(log4cpp::Priority::INFO, 
			 help_messages::getStartMessage().c_str(),
			 false, 
			 true, 
			 debug);
  }
 
  if(argv[optind]==NULL) {
  // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, 
			 "JDL file not specified in the command line arguments. Stop.", 
			 false, 
			 true, 
			 true);
    exit(1);
  } else {
    JDLFile = string(argv[optind]);
    jdlset = true;
  }
  
  if(!ceidset) {
    // ERROR
    string os = string("CEId not specified in the command line arguments. Type ") + argv[0] + " -h for help";
    logger_instance->log(log4cpp::Priority::FATAL, os, false, true, true);
    exit(1);
  }

  if(specifiedDID && automatic_delegation) {
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, "Cannot specify both --autm-delegation and --delegationId options. Stop.", false, true, true);
    exit(1);
  }

  if(!specifiedDID && !automatic_delegation) {
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, 
			 "Must specify one (and only one) of the two options --autm-delegation --delegationId. Stop.", false, true, true);
    exit(1);
  }

  /* ****************************************************************************
   * Create the CREAM proxy client that will handle the communication with
   * the server.
   * Also will try to extract the VO name now
   */ 
  if(!specified_proxyfile) {
    try {
      certfile = cliUtils::getProxyCertificateFile();
    } catch(exception& ex) {
      logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
      exit(1);
    }
  } else
      certfile = proxyfile;
  
  if(debug) 
    logger_instance->log(log4cpp::Priority::DEBUG, string("Using certificate proxy file [")
			 + certfile + "]", true, true, true);

  VOMSWrapper V( certfile, verify_ac_sign  );
  if( !V.IsValid( ) ) {
    
      logger_instance->log(log4cpp::Priority::FATAL, 
			   string("Problems with proxyfile [")+certfile+"]: "+
			   V.getErrorMessage(), true, true, true);
      exit(1);
    
  }
  
  VO_from_cert = V.getVOName( );
  
  
  string VO_JDL = "";
  if(debug) {
    string os = string("VO from certificate=[") + VO_from_cert + "]";
    logger_instance->log(log4cpp::Priority::INFO, os, true, true, !nomsg);
  }
  if(VO_from_cert=="") {
    if(!cli_specified_vo) {
	mustcheckVO = true;
    } else VO_JDL = user_VO;
  } else VO_JDL = VO_from_cert;

  if( VO_from_cert != "" && cli_specified_vo)
  {
    logger_instance->log(log4cpp::Priority::WARN, 
			 string("Specified VO as command line argument ")
			 +"will be overrided by that one from certificate ["+
			 VO_from_cert+"]", true, true, !nomsg);
  }
  
  /* ****************************************************************************
   * Create a jdlHelper object that
   * - checks if the JDL file specified by the user is there
   * - check syntax correctness using WMS.JDL's API
   * - tries to extract VO from JDL if not 
   */
  jdlHelper* jdlH;
  try {
    jdlH = new jdlHelper(JDLFile.c_str(), debug, nomsg);
  } catch(JDLEx& ex) {
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  }
  if(mustcheckVO) {
    if(!jdlH->hasVO()) {
      logger_instance->log(log4cpp::Priority::FATAL,
      			   "VirtualOrganisation not specified in the JDL file. Stop.",
			   false, true, true);
      exit(1);
    }
    VO_JDL = jdlH->getVO();
  }

  /* ****************************************************************************
   * Handles the load of configurations parameters
   */
  vector<string> confiles;
  try {
    confiles = cliUtils::getConfigurationFiles(VO_JDL, 
					       user_conf_file, 
					       DEFAULT_CONF_FILE);
  } catch(no_user_confile_ex& ex) {
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  }
  
  if(debug) {
    vector<string>::const_iterator it;
    for( it = confiles.begin(); it != confiles.end(); it++)
      logger_instance->log(log4cpp::Priority::INFO, 
			   string("Loading configurations from [")
			   +*it+"]...", true, true, true);
  }

  ConfigurationManager confMgr(ConfigurationManager::classad);
  try { confMgr.load(confiles); }
  catch(file_ex& ex) {
    if(debug)
    logger_instance->log(log4cpp::Priority::WARN, 
			 string(ex.what())+
			 ". Using built-in configuration",
			 true, true, !nomsg);
  }

  /* ****************************************************************************
   * Redirect the log/errors/warns on logfile
   */
   if(debug || redir_out) {
     if(!redir_out)
       logfile = logger_instance->getLogFileName(confMgr.getProperty("SUBMIT_LOG_DIR", DEFAULT_LOG_DIR).c_str(), "glite-ce-job-submit");
       cliUtils::mkdir(cliUtils::getPath(logfile));
     logger_instance->setLogFile(logfile.c_str());
   }
     if(debug) {
     string os = string("Logfile is [") + logfile + "]";
     logger_instance->log(log4cpp::Priority::INFO, os, false, false, !nomsg);
   }

  

  /* ****************************************************************************
   *
   * Will check if the CEURL has the correct
   * format that must be host:port/cream-<batchsystem>-<queuename>
   *
   */
  vector<string> ceid_pieces;

  try {
    CEUrl::parseCEID(URL, ceid_pieces, confMgr.getProperty("DEFAULT_CREAM_TCPPORT","8443"));
  } catch(CEUrl::ceid_syntax_ex& ex) {
    // ERROR
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  }
  
  try {
    if(mustcheckVO) // we already checked the JDL contains the VO
      jdlH->process(ceid_pieces[3], ceid_pieces[2], "");
    else // we have to add the user/proxy VO to JDL overwriting it if exists
      jdlH->process(ceid_pieces[3], ceid_pieces[2], VO_JDL);
  } catch(JDLEx& ex) {
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  } catch(exception& ex) {
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  }

  jobmpi = jdlH->isMPIJOB();
  bool isbbu = jdlH->hasISBBU();
  vector<string> ISB = jdlH->getISB();
  vector<string> _ISBBU = jdlH->getISBBU();
  string ISBBU;
  if(_ISBBU.size())
    ISBBU = _ISBBU[0];



  /**
   *
   * Prepare to send ISB via gridftp to the SE returned
   * by CREAM's register remote call
   *
   */
//   int initial_cwdlen=2048;
//   char *cwd, *ret;
//   do {
//     cwd = (char*)malloc(initial_cwdlen);
//     memset((void*)cwd,0,initial_cwdlen);
//     ret = getcwd(cwd, initial_cwdlen);
//     initial_cwdlen *= 2;
//   } while(!ret);
  //string CWD = cwd;

  try {
    jdlH->processISB( fs::current_path( ).string( ), isbbu );
  } catch(exception& ex) {
     logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
     exit(1);
  }
//   vector<string> absolutepath;
//   for(vector<string>::const_iterator isbit = ISB.begin();
//       isbit != ISB.end(); isbit++) 
//     {
//       if(debug)
// 	logger_instance->log(log4cpp::Priority::INFO, string("Processing [")+ *isbit +"]", true, false, true);
//       string thisFile = *isbit;
// 
//             
// 
//       if(boost::starts_with( *isbit, "gsiftp://")) continue; // do not want to do 3rd party copy
//       if(!boost::starts_with( *isbit, "/") && !boost::starts_with( *isbit, "file://") && isbbu) continue;
//       if(boost::starts_with( *isbit, "http://")) continue;
//       if(boost::starts_with( *isbit, "https://")) continue;
// 
//       cout << "Will handle [" << *isbit << "]" << endl;
// 
//       if ( (!boost::starts_with( *isbit, "/")) &&
// 	   (!boost::starts_with( *isbit, "file://")) )
// 	{
// 	  //the file is relative to CWD or to a gridftp URL if specified a OSBBU
// 	  if(!isbbu) {
// 	    absolutepath.push_back(CWD + "/" + *isbit);
// 	    if(debug)
// 	      logger_instance->log(log4cpp::Priority::INFO, string("Added [")+CWD+"/"+ *isbit +"]", true, false, true);
// 	  } else { // GSIFTP
// 	    absolutepath.push_back(ISBBU + "/" + *isbit);
// 	    if(debug)
// 	      logger_instance->log(log4cpp::Priority::INFO, string("Added [")+ISBBU + "/" + *isbit +"]", true, false, true);
// 	  }
// 	} else {
// 	  string thisFile = *isbit;
// 	  if(boost::starts_with( *isbit, "file://"))
// 	    thisFile.replace(0, 7, "");
// 	
// 	  if(debug)
// 	    logger_instance->log(log4cpp::Priority::INFO, string("Adding [")+thisFile+"]", true, false, true);
// 	  absolutepath.push_back(thisFile);
//       }
//     }
//   
//   absolutepath = cliUtils::expandWildcards(absolutepath);
//   
//   try {
// 
//     jdlH->removeWildcardFromISB();
//     jdlH->addFilesToISB(absolutepath);
// 
//   } catch(exception& ex) {
//     logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
//     exit(1);
//   }
  
  /**
   *
   * Checks the complete accesibility (open and read-only) of all files to send
   *
   */
   int fd;
   
   set<string> paths;
   jdlH->getAbsolutePath( paths );
   set<string>::const_iterator path_it = paths.begin();
   
   for( ; path_it != paths.end(); ++path_it) 
     { 
       if((fd=open( (*path_it).c_str(), O_RDONLY|O_LARGEFILE))==-1)
	 {
	   logger_instance->log(log4cpp::Priority::FATAL, "The file [" + *path_it + "] is not there or is not accessible", true, true, true);
	   exit(1);
	 }
       close(fd);
     }

  /**
   *
   * Gets queue name and batch system name
   * from the CEURL command line argument and
   * appends them to the service URL name. Then
   * append the actual path context of the
   * deployed cream service.
   *
   */
   
  string endpoint = ceid_pieces[0];
   
  string serviceAddress = confMgr.getProperty("CREAM_URL_PREFIX", "http://")+
    cliUtils::getCompleteHostName(ceid_pieces[0]) + 
    ":" + 
    ceid_pieces[1]+
    confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX);


  string delService = confMgr.getProperty("CREAMDELEGATION_URL_PREFIX", DEFAULT_PROTO)+
    cliUtils::getCompleteHostName(ceid_pieces[0]) + 
    ":" + 
    ceid_pieces[1]+
    confMgr.getProperty("CREAMDELEGATION_URL_POSTFIX", DEFAULT_CEURLDELEGATION_POSTFIX);

  /**
   *
   * Calls the remote method and handles errors (if any)
   *
   */
   if(fake) return 0;
   int f = -1;

   if(jobid_on_file) {

     bool exists = false;
     if((exists=cliUtils::fileExists(outfile.c_str()))==true) {
	if(!cliUtils::fileIsWritable(outfile.c_str())){
	  logger_instance->log(log4cpp::Priority::FATAL,"Output file exists but is not accessible. Stop", true, true, true);
	  exit(1);
	}
     }
     if(exists) {
       bool isjoblist = cliUtils::isACreamJobListFile(outfile.c_str());
       if(!isjoblist) {
	 if(!noint) {
	   printf("\nWARNING: file [%s] already exists and is not a CREAM job list file. Overwrite it (y/n) ? ", outfile.c_str());
	   char c;
	   cin >> c;
	   if(c!='y') {
	     printf("\nSubmission aborted. Bye!\n");
	     exit(0);
	   }
	 }
	 // Is NOT a JobList => opens file and truncs it
	 try { f=cliUtils::openJobListFile(outfile.c_str(), true); }
	 catch(file_ex& fex) {
		logger_instance->log(log4cpp::Priority::FATAL, fex.what(), true, true, true);
		exit(1);
	 }
       } else {
	  // IS a JobList file => opens file in append mode
	  try { f=cliUtils::appendToJobListFile(outfile.c_str()); }
	  catch(file_ex& fex) {
	    logger_instance->log(log4cpp::Priority::FATAL, fex.what(), true, true, true);
	    exit(1);
	  }
       }
     } else {  
        // DOESN'T EXIST => create file from scratch
	try { f=cliUtils::openJobListFile(outfile.c_str(), false); }
	catch(file_ex& fex) {
	    logger_instance->log(log4cpp::Priority::FATAL, fex.what(), true, true, true);
	    exit(1);
	}
     }
   }// if(jobid_on_file)
   /**
    * Real submission
    *
    */
   string jID;
   string ISB_upload_url;
   int err = 0;
   bool ISB_to_send = false;
   
   if( jdlH->getAbsolutePathSize() )
     ISB_to_send = true;


   string jdlExtras = confMgr.getProperty("JDL_DEFAULT_ATTRIBUTES","");

   if(jdlExtras!="") {
     
      try {
	jdlH->addFrontClassAd(jdlExtras);
      }
      catch(JDLError& ex) {
	logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
	exit(1);
      }
      catch(JDLEx& ex) {
        logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
	exit(1);
      }
      catch(exception& ex) {
	logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
	exit(1);
      }
      catch(...) {
	logger_instance->log(log4cpp::Priority::FATAL, "UNKNOWN ex catched", true, true, true);
	exit(1);
      }
   }

   string JDLBuffer = jdlH->toString();

   // Checking Prologue/Epilogue and PrologueArguments/EpilogueArguments
   //
   if( !jdlH->checkPrologue() )
     {
       logger_instance->log(log4cpp::Priority::FATAL, "Specified 'PrologueArguments' attribute without 'Prologue'", true, true, true);
       exit(1);
     }
   
   if( !jdlH->checkEpilogue() )
     {
       logger_instance->log(log4cpp::Priority::FATAL, "Specified 'EpilogueArguments' attribute without 'Epilogue'", true, true, true);
       exit(1);
     }
   
   try {
     if(debug) {
       string os = string("Registering to [") + serviceAddress + "] JDL=" + JDLBuffer;
       logger_instance->log(log4cpp::Priority::INFO, os, true, true, !nomsg);
     }
     
     if (automatic_delegation)
        DID = make_delegation(delService, certfile, timeout);
     
     string leaseID;
     string delegationProxy;
     leaseID = leaseId;
     JobDescriptionWrapper jd(JDLBuffer, DID, delegationProxy, leaseID, !ISB_to_send, "foo");
     
     map< string, string > properties;string localCreamJID, creamURL;

     AbsCreamProxy::RegisterArrayRequest reqs;
     reqs.push_back( &jd );
     AbsCreamProxy::RegisterArrayResult resp;
     
     AbsCreamProxy *creamClient = CreamProxyFactory::make_CreamProxyRegister( &reqs, &resp, timeout );
     
     if(!creamClient) 
       {
	 logger_instance->log(log4cpp::Priority::FATAL, 
			      string("FAILED TO CREATE AN AbsCreamProxy object!! STOP!!"), 
			      true, 
			      true, 
			      true);
	 return 1;
       }
     boost::scoped_ptr< AbsCreamProxy > tmpRegisterClient;
     tmpRegisterClient.reset( creamClient );

     creamClient->setCredential( certfile );

     creamClient->execute( serviceAddress ); // can raise something
          
     boost::tuple<bool, JobIdWrapper, string> thisRegRes = resp["foo"];
     
     if(JobIdWrapper::OK != thisRegRes.get<0>()) {
       logger_instance->log(log4cpp::Priority::FATAL, thisRegRes.get<2>(), true, true, !nomsg);
       return 1;
     }
     
     creamURL = thisRegRes.get<1>().getCreamURL();
     
     string::size_type loc = creamURL.find(confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX));

     if( loc != string::npos)
       {
	 creamURL = creamURL.substr(0,loc);
       }

      localCreamJID = thisRegRes.get<1>().getCreamJobID();
     jID = creamURL + "/" + localCreamJID; 

     thisRegRes.get<1>().getProperties( properties );

     //Alvise's stuff return 0;

     if(ISB_to_send) {
       ISB_upload_url = properties["CREAMInputSandboxURI"];
       string os = string("JobID=[") + jID + "]";
       if(debug) {
	 logger_instance->log(log4cpp::Priority::INFO, os, true, true, !nomsg);
	 os = string("UploadURL=[") + ISB_upload_url + "]";
	 logger_instance->log(log4cpp::Priority::INFO, os, true, true, !nomsg);
       }
       
       {
       ftpclient ftp( (string("gsiftp://")+endpoint).c_str(), num_streams );

       set<string> absPaths;
       jdlH->getAbsolutePath( absPaths );

       for(set<string>::const_iterator fit = absPaths.begin();
           fit != absPaths.end();
	   fit++) 
       {
         string targetFname = cliUtils::getFileName( (string&)*fit );
	 logger_instance->log(log4cpp::Priority::INFO, string("Sending file [")+ISB_upload_url + "/" + targetFname+"]", true, true, !nomsg);
         if(!ftp.put( fit->c_str( ),  (ISB_upload_url + "/" + targetFname).c_str() ) )
           exit(1);
       }
       } // destroy the ftp object that closes connection with gridftp server
/*       
       for(vector<string>::const_iterator fit = absolutepath.begin();
           fit != absolutepath.end();
	   fit++) 
       {
	 string targetFname = cliUtils::getFileName( (string&)*fit );
	 if(debug)
	   logger_instance->log(log4cpp::Priority::INFO, string("Sending [") + *fit + "] to ["+ISB_upload_url + "/" +
	   targetFname+"]...", true,true, !nomsg);

	 
	 // ************************ VIA GRID FTP globus-url-copy FILE SEND ****************************************
	 
// 	 string os = confMgr.getProperty("GRIDFTP_CLIENT", DEFAULT_GRIDFTP_CLIENT);
// 	 
// 	 os = os + " " + boost::str( boost::format("%d") % num_streams) + " file://" + *fit + " " + ISB_upload_url + "/" + targetFname + " 2>&1";
// 	 if(debug) {
// 	   string log = string("Executing command [") + os + "]";
// 	   logger_instance->log(log4cpp::Priority::INFO, log, true, WRITE_LOG_ON_FILE, !nomsg);
// 	 }
// 	 FILE *res = popen(os.c_str(), "r");
// 	 if(!res) {
// 	   logger_instance->log(log4cpp::Priority::ERROR, "popen failed starting globus-url-copy. Stop.", true, WRITE_LOG_ON_FILE, true);
// 	   exit(1);
// 	 }
// 	 string output;
// 	 while(!feof(res)) {
// 	   output += fgetc(res);
// 	 }
// 	 if(pclose(res)) // the command failed for some reason
// 	   {
// 	     string putlog = string("URL-COPY ERROR OUTPUT: ") + output;
// 	     logger_instance->log(log4cpp::Priority::ERROR, putlog, true, WRITE_LOG_ON_FILE, true);
// 	     exit(1);
// 	   }
	 // ***********************************VIA GRID FTP LIB FILE SEND ******************************************
	
	 if(!getenv("CREAM_CLI_ISB_BYPASS")) {
	   bool ret = ftpclient::put( *fit, ISB_upload_url + "/" + targetFname, 4);
	   if(!ret) {
	     logger_instance->log(log4cpp::Priority::FATAL, string("Error sending file [")+*fit+"]", true, true, true);
	     exit(1);
	   }
	 }
	 // ********************************************************************************************************
       }
*/
       // Now proceed with JOBSTART
       if(debug)
	 logger_instance->log(log4cpp::Priority::INFO, "Now invoking JobStart for JobID [" + jID + "]", true, true, !nomsg);
       
       //JobPropertyWrapper jpw;
       vector<JobPropertyWrapper> jpw_vector;
       //jpw_vector.push_back( jpw );
       JobIdWrapper jiw(localCreamJID, creamURL, jpw_vector);
       vector< JobIdWrapper > param1;
       param1.push_back( jiw );
       
       JobFilterWrapper jfw( param1, vector<string>(), -1, -1, DID, leaseID );
       
       ResultWrapper result;

       //AbsCreamProxy* creamClientStart = CreamProxyFactory::make_creamproxy( AbsCreamProxy::START, 30 );
       AbsCreamProxy* creamClientStart = CreamProxyFactory::make_CreamProxyStart( &jfw, &result, timeout );
       if(!creamClientStart)
	 {
	   logger_instance->log(log4cpp::Priority::FATAL, "FAILED CREATION OF AN AbsCreamProxy object! STOP!!", true, true, true);
	   return 1;
	 }
       boost::scoped_ptr< AbsCreamProxy > tmpClient;
       tmpClient.reset( creamClientStart );
       creamClientStart->setCredential( certfile );
       // creamClientStart->setRequestParameters( &jfw, &result);
       creamClientStart->execute( serviceAddress );
       // creamClientStart->garbage_collect();
     } // if(ISB_to_send)
    } catch(BaseException& ex) {
      jID = ex.what();
      err = 1;
    } catch( exception& ex )  {
      jID = ex.what( );
      err = 1;
    } 

   if(jobid_on_file) {
     if(!err) {
       try { cliUtils::writeJobID(f, jID); }
       catch(file_ex& fex) {
	  logger_instance->log(log4cpp::Priority::FATAL, fex.what(), true, true, true);
	  close(f);
	  exit(1);
       }
     }
     close(f);
   }

   printf("%s\n", jID.c_str()); // print on STDOUT even if --output has been specified

   //delete creamClient;
   
   return err;
}

/**
 *
 * Streaming call backs implementations
 *
 */
// void *dime_read_open(struct soap *soap, void *handle, const char *id, const char *type, const char *options)
// {
//   return handle;
// }
// 
// void dime_read_close(struct soap *soap, void *handle)
// {
//   fclose((FILE*)handle);
// }
// 
// size_t dime_read(struct soap *soap, void *handle, char *buf, size_t len)
// {
//   return fread(buf, 1, len, (FILE*)handle);
// }
// 
void printhelp() {
  printf("%s\n",help_messages::getSubmitHelp().c_str());
}

string make_delegation( const string& deleg_service, const string& certfile, const int timeout) throw()
{
  creamApiLogger* logger_instance = creamApiLogger::instance();
  string delegID;
  glite::ce::cream_client_api::certUtil::generateUniqueID( delegID );
  AbsCreamProxy *creamDelegClient =  CreamProxyFactory::make_CreamProxyDelegate( delegID, timeout );

    //AbsCreamProxy *creamDelegClient = CreamProxyFactory::make_creamproxy(AbsCreamProxy::DELEGATE, 30);

  if(!creamDelegClient)
    {
      logger_instance->log(log4cpp::Priority::FATAL, "FAILED CREATION OF AN AbsCreamProxy object! STOP!!", true, true, true);
      exit( 1 );
    }

  boost::scoped_ptr< AbsCreamProxy > tmpClient;
  tmpClient.reset( creamDelegClient );

  creamDelegClient->setCredential( certfile );

  try {
    creamDelegClient->execute( deleg_service );
    //creamDelegClient->garbage_collect();
  } catch(exception& ex) {
    logger_instance->log(log4cpp::Priority::FATAL, ex.what(), true, true, true);
    exit(1);
  }
  
  return delegID;
}
